#include<math.h>
#include<stdlib.h>
#include<stdio.h>

//#define INTERIOR       0
//#define FRONTERA_REAL  1

#include "libs/jag_mem.cpp"
#include "libs/jag_phi.cpp"
#include "libs/jag_vector.cpp"
#include "libs/jag_cond.cpp"
#include "libs/qSort.cpp"

#include "edp_ut.cpp"

#include "varios.cpp"

#include "libs/tree2d.cpp"
#include "libs/interpola_tps2d.cpp"


//-------------------------------------------------------------------
void get_error_localIF(tree **data, int N, double &maxi)
// Determino el error de interpolacion local de los nodos interiores y de frontera..es decir, de todos...jeje
{
  register int     i,j;
  int     K=0;
  tree    *seleccionado;
  tree    *Vecinos[128]; 
  double  X[128],Y[128],F[128];
  double  xc,yc,fc;
  double  val;

//Limpio los errores de interpolacion local
 for(i=0;i<N;i++)
 {
   data[i]->error = 0.0;
 }

//Interpolo al estilo de Iskee
 for(i=0;i<N;i++)
 {
  //Aqui debo de buscar a sus vecinos mas cercanos para calcular el error
  //de interpolacion local
     seleccionado = data[i];
     if(seleccionado==NULL){
       printf("ERROR-2: el nodo %d no se encontro en archivo = '%s' , linea = %d.\n",i,__FILE__,__LINE__);
     }
   else{

        K=0;
        xc = seleccionado->x;
        yc = seleccionado->y;
        fc = seleccionado->sol;

      //Selecciono los K-vecinos
        get_kvecinos2(seleccionado,Vecinos,K);

       if(K<=2){
            printf("Esta raro:  K = %d en (%f,%f)\n",K,xc,yc);
           exit(1);
      }

        for( j=0; j<K; j++)
        {
           X[j]=Vecinos[j]->x;
           Y[j]=Vecinos[j]->y;
           F[j]=Vecinos[j]->sol;
        }
          double aux;
          val = interpola_tps2d_P1(xc,yc,X,Y,F,K);  // Phi + 1 + x + y
          val = fabs(fc-val);
          seleccionado->error = val;
   }

 }

//Determino el error maximo de interpolacion local
  maxi=0;
 for(i=0;i<N;i++)
 {
  if(data[i]->error>maxi)  maxi = data[i]->error;
 }
 printf("Error local maximo = %f.\n",maxi);
}


//-------------------------------------------------------------------
int isin(tree *nodo, tree **nodos, int N)
{
 if(N<=0) return 0;

 for(int i=0;i<N;i++)
 {
   if(nodo==nodos[i])
   {
     return 1;
   }
 }
 return 0;
}


//-------------------------------------------------------------------
void   refinaIF(tree *raiz,tree **data, int N, int min_nivel_ref,int max_nivel_ref,
                 double thetar, double thetac,int &ref, int &rem)
{
 tree **papa_refinados;
 tree  *seleccionado;
 int    cont=0;
 register  int    i;
 ref = 0; rem = 0;
 papa_refinados = (tree **)malloc(5*N*sizeof(tree));


//Cuento cuantos nodos van a pasar el refinamiento local
 for(i=0;i<N;i++)
 {
   if(data[i]->error>=thetar)
   {
      ref=ref+1;
   }
   if(data[i]->error<thetac)
   {
      rem=rem+1;
   }
 }// for i

 printf("Posibles candidatos a refinar = %d de %d\n",ref,N);
 printf("Posibles candidatos a remover = %d de %d\n",rem,N);


 int candidatos;
 candidatos = ref;



 ref=0;
//Refino los nodos
//for(int iter=0;iter<3;iter++)
 for(i=0;i<N;i++)
 {
   seleccionado = data[i];
   if(seleccionado->error>=thetar)
   {
     if(refina(seleccionado,max_nivel_ref)==1)
     {
        ref=ref+1;
        papa_refinados[cont++] = seleccionado->papa;
     }
   }
 }// for i


 rem=0;

//Remuevo los nodos
 for(i=0;i<N;i++)
 {
   seleccionado = data[i];
   if(seleccionado->error<=thetac)
    if(isin(seleccionado->papa,papa_refinados,cont)==0)
      if(remueve2(seleccionado,min_nivel_ref)==1)
      {
         rem=rem+1;
         papa_refinados[cont++] = seleccionado->papa;
      }
 }// for i


 free(papa_refinados);
}

//-------------------------------------------------------------------
void restriccion_suavidad2(tree *raiz, int &restriccion, int max_nivel_ref)
//Checo la restriccion de suavidad 2:1
{
   tree *Vecinos[128];
   double *x, *y;
   tree   **data;
   int   K,COTA;
   int   Ni,N=0;

   cont_hojas(raiz,N);

   x=new double[N]; y=new double[N];
   data = (tree **)malloc(N*sizeof(tree));

   Ni=0;  get_datos_interior(raiz,x,y,data,Ni);
   N=Ni;  get_datos_frontera(raiz,x,y,data,N);

   delete[] x; delete[] y;

   restriccion=0;
for(int i=0;i<N;i++)
{
  tree *select;

  select = data[i];

  K=0;
  get_kvecinos2(select,Vecinos,K);

  for(int j=0;j<K;j++)
  {
      if(fabs(double(select->nivel)-double(Vecinos[j]->nivel))>1.0)
      {
        if(select->nivel>Vecinos[j]->nivel){
           if(refina(Vecinos[j],max_nivel_ref)==1)
             restriccion++;
        }
        else{
           if(refina(select,max_nivel_ref)==1)
             restriccion++;
         }
      }
   }      // if der
}
 free(data);
}

//--------------------------------------------------------------------
void reconstruye(double *x,double *y, double *lambda,double c,int N, double *sol_numerica)
{
int     i,j;
double xc,yc;
double s;

 for(i=0;i<N;i++)
 {
  s=0.0;
   for(j=0;j<N;j++)
   {
      s += lambda[j]*phitps42(x[i]-x[j],y[i]-y[j],0,0);
   }

   s = s + lambda[N+0];
   s = s + lambda[N+1]*x[i];
   s = s + lambda[N+2]*y[i];
   s = s + lambda[N+3]*x[i]*y[i];
   s = s + lambda[N+4]*x[i]*x[i];
   s = s + lambda[N+5]*y[i]*y[i];

   sol_numerica[i] = s;
 }
}

//--------------------------------------------------------------------
double interpola(double xc,double yc, double *x,double *y, double *lambda,double c,int N)
{
 int      j;
double s;

  s=0.0;
   for(j=0;j<N;j++)
   {
      s += lambda[j]*phitps42(xc-x[j],yc-y[j],0,0);
   }

   s = s + lambda[N+0];
   s = s + lambda[N+1]*xc;
   s = s + lambda[N+2]*yc;
   s = s + lambda[N+3]*xc*yc;
   s = s + lambda[N+4]*xc*xc;
   s = s + lambda[N+5]*yc*yc;

 return s;
 
}

//--------------------------------------------------------------------
//Proyeccion
void mult_matrix_vec(double **P, double *lambda,int m,double *z)
{
 double s;	
  for(int i=0;i<m;i++)	
  {
	  s = 0.0; 
   for(int j=0;j<m;j++)	
   {
	 s = s + P[i][j]*lambda[j];
   }	     
  s = s + P[i][m+0]*lambda[m+0];
  s = s + P[i][m+1]*lambda[m+1];
  s = s + P[i][m+2]*lambda[m+2];
  s = s + P[i][m+3]*lambda[m+3];
  s = s + P[i][m+4]*lambda[m+4];
  s = s + P[i][m+5]*lambda[m+5];

  z[i] = s; 
 } 
      
}
//--------------------------------------------------------------------
//Correcion
void corrige_bc(double *x,double *y, int ni, int m, double t,double *f)
{
  for(int i=ni;i<m;i++)		
     f[i] = u_t(x[i],y[i],t);
}
//--------------------------------------------------------------------
double* resuelve_IC(double *x,double *y, int N)
{
 double **A;
 double  *f;
 double  *lambda;
 int     *pivot_index;
 double   cond;
 int      Pol=6;

//Reservo la memoria dinamica
  A           = make_dmatrix(N+Pol,N+Pol);
  f           = make_dvector(N+Pol);
  lambda      = make_dvector(N+Pol);
  pivot_index = make_ivector(N+Pol);

//Relleno de zeros
  fill(f,N+Pol,0.0);

 for(int i=0;i<(N+Pol);i++)
   fill(A[i],N+Pol,0.0);

//Creo la matriz de interpolacion  
for(int i=0;i<N;i++)
{
    for(int j=0;j<N;j++)
    {
        A[i][j] = phitps42(x[i]-x[j],y[i]-y[j],0,0);
    }
  f[i]=u_t0(x[i],y[i]); 

 A[i][N+0] = 1.0;
 A[i][N+1] = x[i];
 A[i][N+2] = y[i];
 A[i][N+3] = x[i]*y[i];
 A[i][N+4] = x[i]*x[i];
 A[i][N+5] = y[i]*y[i];

 A[N+0][i] = 1.0;
 A[N+1][i] = x[i];
 A[N+2][i] = y[i];
 A[N+3][i] = x[i]*y[i];
 A[N+4][i] = x[i]*x[i];
 A[N+5][i] = y[i]*y[i];

}

//Factorizo la matriz
   Factor(A,N+Pol,&cond,pivot_index);

   printf("num.condi   = %e\n",cond);  

//Resuelvo mediante LU
  copy(f,N+Pol,lambda);
  Solve(A,N+Pol,pivot_index,lambda);

//Libero la memoria dinamica
  free_dmatrix(A,N+Pol);
  free_dvector(f);
  free_ivector(pivot_index);

 return lambda;
}

//-------------------------------------------------------------------------------------------------------
//Creo la matriz-P1
void crea_P1(double *x,double *y, int N,double stept,double **P)
 {
double u,ux,uy;

  for(int i=0;i<N;i++)
  {
   for(int j=0;j<N;j++)
   {
	  u  = phitps42(x[i]-x[j],y[i]-y[j],0,0);   
	 ux  = phitps42(x[i]-x[j],y[i]-y[j],0,1);   
	 uy  = phitps42(x[i]-x[j],y[i]-y[j],0,3);   	 
     P[i][j] = u - 0.5*stept*(ux+uy);
   } 
    P[i][N+0] = 1.0       - 0.5*stept*0.0      - 0.5*stept*0.0;
    P[i][N+1] = x[i]      - 0.5*stept*1.0      - 0.5*stept*0.0;
    P[i][N+2] = y[i]      - 0.5*stept*0.0      - 0.5*stept*1.0;
    P[i][N+3] = x[i]*y[i] - 0.5*stept*y[i]     - 0.5*stept*x[i];
    P[i][N+4] = x[i]*x[i] - 0.5*stept*2.0*x[i] - 0.5*stept*0.0;
    P[i][N+5] = y[i]*y[i] - 0.5*stept*0.0      - 0.5*stept*2.0*y[i];
  }
}
//-------------------------------------------------------------------------------------------------------
//Creo la matriz-P2 
void crea_P2(double *x,double *y,  int N,double stept,double **P)
 {
double u,ux,uy;

  for(int i=0;i<N;i++)
  {
   for(int j=0;j<N;j++)
   {
	 ux  = phitps42(x[i]-x[j],y[i]-y[j],0,1);   
	 uy  = phitps42(x[i]-x[j],y[i]-y[j],0,3);   	 
     P[i][j] =  -0.5*stept*(ux+uy);
   } 
    P[i][N+0] = - 0.5*stept*0.0      - 0.5*stept*0.0;
    P[i][N+1] = - 0.5*stept*1.0      - 0.5*stept*0.0;
    P[i][N+2] = - 0.5*stept*0.0      - 0.5*stept*1.0;
    P[i][N+3] = - 0.5*stept*y[i]     - 0.5*stept*x[i];
    P[i][N+4] = - 0.5*stept*2.0*x[i] - 0.5*stept*0.0;
    P[i][N+5] = - 0.5*stept*0.0      - 0.5*stept*2.0*y[i];
  }
}

//-------------------------------------------------------------------------------------------------------
//Creo la matriz-M
void crea_M(double *x,double *y,int ni, int N,double stept,double **M)
 {
 double u,ux,uy;

 for(int i=0;i<(N+6);i++)
   fill(M[i],N+6,0.0);

  for(int i=0;i<ni;i++)
  {
   for(int j=0;j<N;j++)
   {
	  u = phitps42(x[i]-x[j],y[i]-y[j],0,0);   
	 ux = phitps42(x[i]-x[j],y[i]-y[j],0,1);   
	 uy = phitps42(x[i]-x[j],y[i]-y[j],0,3);   	 
     M[i][j] = u + 0.5*stept*(ux+uy);
   } 
    M[i][N+0] = 1.0  + 0.5*stept*0.0 + 0.5*stept*0.0;
    M[i][N+1] = x[i] + 0.5*stept*1.0 + 0.5*stept*0.0;
    M[i][N+2] = y[i] + 0.5*stept*0.0 + 0.5*stept*1.0;
    M[i][N+3] = x[i]*y[i] + 0.5*stept*y[i] + 0.5*stept*x[i];
    M[i][N+4] = x[i]*x[i] + 0.5*stept*2.0*x[i] + 0.5*stept*0.0;
    M[i][N+5] = y[i]*y[i] + 0.5*stept*0.0 + 0.5*stept*2.0*y[i];

    M[N+0][i] = 1.0;
    M[N+1][i] = x[i];
    M[N+2][i] = y[i];
    M[N+3][i] = x[i]*y[i];
    M[N+4][i] = x[i]*x[i];
    M[N+5][i] = y[i]*y[i];
  }

  for(int i=ni;i<N;i++)
  {
   for(int j=0;j<N;j++)
   {
	  u = phitps42(x[i]-x[j],y[i]-y[j],0,0);   
     M[i][j] = u;
   } 
    M[i][N+0] = 1.0;
    M[i][N+1] = x[i];
    M[i][N+2] = y[i];
    M[i][N+3] = x[i]*y[i];
    M[i][N+4] = x[i]*x[i];
    M[i][N+5] = y[i]*y[i];

    M[N+0][i] = 1.0;
    M[N+1][i] = x[i];
    M[N+2][i] = y[i];
    M[N+3][i] = x[i]*y[i];
    M[N+4][i] = x[i]*x[i];
    M[N+5][i] = y[i]*y[i];
  }
}

//-------------------------------------------------------------------------------------------------------
//Creo la matriz-A
void crea_A(double *x,double *y, int N,double **A,double *f)
 {
 double u,ux,uy;

  fill(f,N+6,0.0);

  for(int i=0;i<(N+6);i++)
    fill(A[i],N+6,0.0);

for(int i=0;i<N;i++)
{
    for(int j=0;j<N;j++)
    {
        A[i][j] = phitps42(x[i]-x[j],y[i]-y[j],0,0);
    }
  f[i]=u_t0(x[i],y[i]); 

 A[i][N+0] = 1.0;
 A[i][N+1] = x[i];
 A[i][N+2] = y[i];
 A[i][N+3] = x[i]*y[i];
 A[i][N+4] = x[i]*x[i];
 A[i][N+5] = y[i]*y[i];

 A[N+0][i] = 1.0;
 A[N+1][i] = x[i];
 A[N+2][i] = y[i];
 A[N+3][i] = x[i]*y[i];
 A[N+4][i] = x[i]*x[i];
 A[N+5][i] = y[i]*y[i];

}

}
//--------------------------------------------------------------------
int main(int argc, char **argv)
{
 double  t,tini,tmax,stept;	
 int     n,N,ni;

 
 double *sol_numerica;
 double *sol_analitica;
 double *sol_analitica_numerica; 
 
 double *x,*y; 
 double *lambdaIC;

 
 double  e_rms,e_inf;
 double xmin,xmax,ymin,ymax;
 mesh   malla;   

 int    max_nivel_ref;
 int    min_nivel_ref; 
 int    nivel_ref;
 double maxi_elocal;
 double thetar,thetac;
 tree    *raiz;
 tree   **data;


 if(argc!=6){
   printf("ERROR:  requiere  <min_nivel> <thetar> <thetac> <stept> <tmax>\n\n");
   exit(1);	 
 }
 
nivel_ref     = atoi(argv[1]);
thetar        = atof(argv[2]);  
thetac        = atof(argv[3]);  
stept         = atof(argv[4]);  
tmax          = atof(argv[5]);  

 tini = 0.0;


 max_nivel_ref = 15;
 min_nivel_ref = 3;   // 128 nodos


 printf("-----------------------------------------------------\n");
 printf("nivel_ref = %d, total = %f\n",nivel_ref,pow(4,nivel_ref));
 printf("thetar        = %f\n",thetar);   
 printf("thetac        = %f\n",thetac);    



//Creo el grid cartesiano
 xmin =  0;
 ymin =  0;
 xmax =  1;
 ymax =  1;

//Creo la raiz inicial del arbol quad-tree
  raiz = crea_raiz(xmin,xmax,ymin,ymax);

//Asigno al parametro-c inicial
   raiz->c = 4;


//Refino todas las hojas varias veces--jeje
  for(int i=0;i<nivel_ref;i++)
  {
    refina_todo(raiz,max_nivel_ref); clean_band(raiz);
  }

//Aqi viene la opcion de refinar la condicion inicial
for(int iter=0;iter<50;iter++)
{
   clean_frontera(raiz);
   marca_frontera(raiz);

// Ajusto las fronteras
 //  ajusta_fronteras(raiz,xmin,xmax,ymin,ymax);

//Obtengo los nodos tipo frontera e interiores
   N=0; cont_hojas(raiz,N);

   printf("------------------------------------ iter=%d, N=%d\n",iter,N);

   x    = make_dvector(N); 
   y    = make_dvector(N);
   data = (tree **)malloc(N*sizeof(tree));   

   ni=0; N=0;
   get_datos_interior(raiz,x,y,data,ni);
   N=ni;
   get_datos_frontera(raiz,x,y,data,N);
   
   sol_analitica          = make_dvector(N);
   sol_numerica           = make_dvector(N);
   sol_analitica_numerica = make_dvector(N);   

   printf("ni=%d    n=%d\n",ni,N);


   lambdaIC = resuelve_IC(x,y,N);

   reconstruye(x,y,lambdaIC,0,N,sol_numerica);

   for(int i=0;i<N;i++)
     data[i]->sol = sol_numerica[i];

   get_error_localIF(data,N,maxi_elocal);
 
//Calculo el error cuadratico medio y el error maximo
//sobre el mismo grid
  for(int i=0;i<N;i++)   sol_analitica[i] = u_t0(x[i],y[i]);

  save_gnu_solucion("real",x,y,sol_analitica,N);  
  save_gnu_solucion("numerica",x,y,sol_numerica,N);  

 //Refino y remuevo
  if(maxi_elocal<=thetar)
  {
    printf("-----------------------------------------------------------\n");
    printf("Se alcanzo el error de interpolacion local: %f<%f.\n",maxi_elocal,thetar);
    printf("-----------------------------------------------------------\n\n");
    iter=500;
  }
 if(iter<400)
 {
//Refino los condenados nodos
   int ref,rem;
   ref=0;   rem=0;
   refinaIF(raiz,data,N,min_nivel_ref,max_nivel_ref, thetar,thetac,ref,rem);
   printf("Refinados = %d , Removidos = %d,  newN = %d\n",ref,rem,N+3*ref-3*rem);

//Aplico la restriccion de suavidad
 for(int pp=0;pp<5;pp++)
 {
  int restriccion;
  restriccion=0;
  restriccion_suavidad2(raiz,restriccion,max_nivel_ref);
  if(restriccion>0)  printf("Restriccion 2:1 se agregaron= %d \n",restriccion);
 }
}
   free_dvector(x); 
   free_dvector(y);
   free_dvector(lambdaIC);
   free(data);


   free_dvector(sol_analitica);
   free_dvector(sol_numerica);
   free_dvector(sol_analitica_numerica);   
}//iter


  printf("bye.\n");

//Ya termine RLN para la condicion inicial, ahora viene la parte temporal
  double **A,**M,**P1,**P2;
  double  *lambda1,*lambda2;
  int          *pivot_indexM;
  int          *pivot_indexA;
  double  *f1,*f2;
  double  cond;
  double  u,ux,uy;

//Obtengo los datos
   N=0; cont_hojas(raiz,N);
   x        = make_dvector(N);    y       = make_dvector(N);
   data = (tree **)malloc(N*sizeof(tree));   
   ni=0; N=0;   get_datos_interior(raiz,x,y,data,ni);
   N=ni;            get_datos_frontera(raiz,x,y,data,N);

//Reservo la memoria dinamica
  A                      = make_dmatrix(N+6,N+6);
  M                      = make_dmatrix(N+6,N+6);
  P1                    = make_dmatrix(N+6,N+6);
  P2                    = make_dmatrix(N+6,N+6);
  f1                      = make_dvector(N+6);
  f2                      = make_dvector(N+6);
  lambda1         = make_dvector(N+6);
  lambda2         = make_dvector(N+6);
  pivot_indexM = make_ivector(N+6);
  pivot_indexA = make_ivector(N+6);

  sol_analitica                    = make_dvector(N);
  sol_numerica                  = make_dvector(N);
  sol_analitica_numerica = make_dvector(N);


//Creo la matriz de interpolacion  
  crea_A(x,y,N,A,f1);

//Factorizo la matriz
   Factor(A,N+6,&cond,pivot_indexA);
   printf("num.condi   = %e\n",cond);  
 
//Resuelvo mediante LU
  copy(f1,N+6,lambda1);
  Solve(A,N+6,pivot_indexA,lambda1);
 

//Calculo el error cuadratico medio y el error maximo
//sobre el mismo grid
  for(int i=0;i<N;i++)
    sol_analitica[i] = u_t0(x[i],y[i]);

  reconstruye(x,y,lambda1,0,N,sol_numerica);

  e_rms = calc_erms(sol_analitica,sol_numerica,N);
  e_inf = calc_einf(sol_analitica,sol_numerica,N);
  
  for(int i=0;i<N;i++) 
      sol_analitica_numerica[i] = sol_analitica[i]-sol_numerica[i];     
    
  save_gnu_solucion("numerica",x,y,sol_numerica,N);  
  save_gnu_solucion("real",x,y,sol_analitica,N);     
  save_gnu_solucion("realnumerica",x,y,sol_analitica_numerica,N);      


 printf("t=%f, e_rms       = %e\n",0.0,e_rms);
 printf("t=%f, e_inf       = %e\n",0.0,e_inf); 


//Creo la matriz-P1
  crea_P1(x,y,N,stept,P1);

//Creo la matriz-P2
  crea_P2(x,y,ni,stept,P2);

//Creo la matriz-M  
 crea_M(x,y,ni, N,stept,M);

//Factorizo la matriz-M
   Factor(M,N+6,&cond,pivot_indexM);

   printf("num.condi   = %e\n",cond);  
   

//Itero sobre el tiempo   
  t = tini;
  
  int ask = int( (tmax/stept)/200.0);
  int cont=0;

 while(t<tmax)
 {
    //printf("---------------------------------------------------- time = %f , N=%d\n",t,N);
 
    //Proyeccion
       fill(f1,N+6,0.0);
       mult_matrix_vec(P1,lambda1,N,f1);
      //Correcion
       corrige_bc(x,y,ni,N,t+stept,f1);      
    //Solucion	 
       copy(f1,N+6,lambda2);       Solve(M,N+6,pivot_indexM,lambda2);

      //Proyeccion
       fill(f2,N+6,0.0);
      mult_matrix_vec(P2,lambda2,N,f2);

      add_vector(f1,f2,N,f1);

    //Correcion
       corrige_bc(x,y,ni,N,t+stept,f1);    
  
    //Solucion	 
       copy(f1,N+6,lambda1);       Solve(A,N+6,pivot_indexA,lambda1);  
  
    //Avanzo en el tiempo   
       t = t + stept; 

          free_dvector(sol_numerica);
          sol_numerica = make_dvector(N);
      //Reconstruyo la solucion 
         reconstruye(x,y,lambda1,0,N,sol_numerica);
         char filename[255];
         sprintf(filename,"numerica%d",cont);
        save_gnu_solucion(filename,x,y,sol_numerica,N);  

      cont++;
   if(!(cont%5))
   {
          free_dvector(sol_numerica);
          sol_numerica = make_dvector(N);
      //Reconstruyo la solucion 
         reconstruye(x,y,lambda1,0,N,sol_numerica);
      //Asigno la solucion 
        for(int i=0;i<N;i++)  data[i]->sol = sol_numerica[i];
      //Obtengo el error de interpolacion local 
        get_error_localIF(data,N,maxi_elocal);

         double *tmp = make_dvector(N);
         for(int i=0;i<N;i++)           tmp[i] = data[i]->error;
        save_gnu_solucion("errorlocalt",x,y,tmp,N);  
         free_dvector(tmp); 
          

      //Pregunto si se rebaso el error de interpolacion local
       if(maxi_elocal>thetar)
       { 
           printf("time = %f  N = %d\n",t,N);
            //Refino/Remuevo creando el nuevo grid
               int ref,rem;           ref=0;   rem=0;
               refinaIF(raiz,data,N,min_nivel_ref,max_nivel_ref, thetar,thetac,ref,rem);
           //Aplico la restriccion de suavidad
               for(int pp=0;pp<5;pp++)
               {
                int restri;  restri=0;
                restriccion_suavidad2(raiz,restri,max_nivel_ref);
                //if(restri>0) printf("Restriccion 2:1 se agregaron= %d\n",restri);
               }

               clean_frontera(raiz);   marca_frontera(raiz);

              //En el nuevo grid interpolo con respecto de lambda
                  int N2,ni2;
                  double *x2,*y2;
                  double *H;
                   tree   **data2;

               N2=0;  cont_hojas(raiz,N2);
                x2       = make_dvector(N2); 
                y2       = make_dvector(N2);
                H        = make_dvector(N2+6);
                data2 = (tree **)malloc(N2*sizeof(tree));   

                 ni2=0; N2=0;
                 get_datos_interior(raiz,x2,y2,data2,ni2);
                  N2=ni2;
                  get_datos_frontera(raiz,x2,y2,data2,N2);

          //Interpolo la solucion en  el nuevo grid para encontrar la lambda
              fill(H,N2+6,0.0);
              for(int i=0;i<N2;i++)  
                        H[i] =  interpola(x2[i],y2[i],x,y,lambda1,0,N);

          //Libero la memoria dinamica anterior
            free_dvector(x);
            free_dvector(y);
            free_dvector(lambda1);
            free_dvector(lambda2);
            free_ivector(pivot_indexA);
            free_ivector(pivot_indexM);
            free_dvector(f1);
            free_dvector(f2);
            free_dmatrix(P1,N+6);
            free_dmatrix(P2,N+6);
            free_dmatrix(A,N+6);
            free_dmatrix(M,N+6);

      //Creo la nueva memoria dinamica
              N = N2;
             ni  = ni2;
              x = x2; 
              y = y2;
             lambda1 = make_dvector(N+6);
             lambda2 = make_dvector(N+6);
             f1              = make_dvector(N+6);
             f2              = make_dvector(N+6);
            M             = make_dmatrix(N+6,N+6);
            P1             = make_dmatrix(N+6,N+6);
            P2             = make_dmatrix(N+6,N+6);
            A             = make_dmatrix(N+6,N+6);
            pivot_indexA     = make_ivector(N+6);
            pivot_indexM     = make_ivector(N+6);

       //Creo las nuevas matrices
             crea_A(x,y,N,A,f1);
             crea_P1(x,y,N,stept,P1);
             crea_P2(x,y,ni,stept,P2);
             crea_M(x,y,ni, N,stept,M);

         //Factorizo la matriz-M y A
                 Factor(A,N+6,&cond,pivot_indexA);
                 printf("num.condi   = %e\n",cond);  

                 Factor(M,N+6,&cond,pivot_indexM);
                 printf("num.condi   = %e\n",cond);  

         //Determino la nueva lambda
            copy(H,N+6,lambda1);
            printf("LINE = %d\n",__LINE__);
            Solve(A,N+6,pivot_indexA,lambda1);
            printf("LINE = %d\n",__LINE__);
            free(data);
            free_dvector(H);
            printf("LINE = %d\n",__LINE__);
            data = data2;   
            printf("LINE = %d\n",__LINE__);
       }//if(maxi_elocal>thetar)
   }//if(!(cont%ask))
 }// while(t<tmax)
  
  printf("\n");
  
//Calculo el error cuadratico medio y el error maximo
//sobre el mismo grid
  for(int i=0;i<N;i++)
    sol_analitica[i] = u_t(x[i],y[i],t);

  reconstruye(x,y,lambda1,0,N,sol_numerica);

  e_rms = calc_erms(sol_analitica,sol_numerica,N);
  e_inf = calc_einf(sol_analitica,sol_numerica,N);
  
  for(int i=0;i<N;i++)     
     sol_analitica_numerica[i] = sol_analitica[i]-sol_numerica[i];     
    
  save_gnu_solucion("numericat",x,y,sol_numerica,N);  
  save_gnu_solucion("realt",x,y,sol_analitica,N);     
  save_gnu_solucion("realnumericat",x,y,sol_analitica_numerica,N);      

 printf("-----------------------------------------------------\n");     
 printf("Advection - Implicit - RLN - TPS\n");
 printf("thetar      = %f\n",thetar);   
 printf("thetac      = %f\n",thetac);    
 printf("stept       = %f\n",stept);   
 printf("tmax        = %f\n",tmax);    
 printf("t=%f, e_rms       = %f\n",t,e_rms);
 printf("t=%f, e_inf       = %f\n",t,e_inf);
 printf("-----------------------------------------------------\n");     
 

return 0;
}

