/*
  This software is under the GNU license.

  Jose Antonio Munioz Gomez
     jantonio@inaoep.mx
       2004 -  INAOEP
*/
#ifndef _JAG_MEM_CPP_
#define _JAG_MEM_CPP_


#include <stdio.h>
#include <stdlib.h>
#include "jag_mem.h"

/*---------------------------------------------------------------------------*/

void msalida(void)
{
   fflush(NULL);
   #ifdef MPI
      MPI_Finalize();
   #else
      exit(EXIT_FAILURE);
   #endif
}

/*---------------------------------------------------------------------------*/

double **make_dmatrix_lf(int m, int n,int linea, char *file)
{
 int  i;
 double **x;

 if(n<0 || m<0){
        fprintf(stderr,"?? ERROR ==> Archivo '%s' linea '%d'\n",file,linea);
	fprintf(stderr,"             La memoria no puede ser negativa '%d %d'.\n",m,n);
	msalida();
 }

 x= (double **)malloc(m*sizeof(double));

 if(x==NULL){
        fprintf(stderr,"?? ERROR ==> Archivo '%s' linea '%d'\n",file,linea);
	fprintf(stderr,"             Memoria Dinamica no disponible.\n");
	msalida();
 }

 for(i=0;i<m;i++)
 {
   x[i]=(double *)malloc(n*sizeof(double));
   if(x[i]==NULL){
        fprintf(stderr,"?? ERROR ==> Archivo '%s' linea '%d'\n",file,linea);
	fprintf(stderr,"             Memoria Dinamica no disponible.\n");
	msalida();
   }
 }
 return x;
}

/*---------------------------------------------------------------------------*/

float **make_smatrix_lf(int m, int n,int linea, char *file)
{
 int     i;
 float **x;

 if(n<0 || m<0){
        fprintf(stderr,"?? ERROR ==> Archivo '%s' linea '%d'\n",file,linea);
	fprintf(stderr,"             La memoria no puede ser negativa '%d %d'.\n",m,n);
	msalida();
 }

 x= (float **)malloc(m*sizeof(float));

 if(x==NULL){
        fprintf(stderr,"?? ERROR ==> Archivo '%s' linea '%d'\n",file,linea);
	fprintf(stderr,"             Memoria Dinamica no disponible.\n");
	msalida();
 }

 for(i=0;i<m;i++)
 {
   x[i]=(float *)malloc(n*sizeof(float));
   if(x[i]==NULL)
   {
        fprintf(stderr,"?? ERROR ==> Archivo '%s' linea '%d'\n",file,linea);
	fprintf(stderr,"             Memoria Dinamica no disponible.\n");
	msalida();
   }
 }
 return x;
}

/*---------------------------------------------------------------------------*/

int **make_imatrix_lf(int m, int n,int linea, char *file)
{
 int  i;
 int **x;

 if(n<0 || m<0){
        fprintf(stderr,"?? ERROR ==> Archivo '%s' linea '%d'\n",file,linea);
	fprintf(stderr,"             La memoria no puede ser negativa '%d %d'.\n",m,n);
	msalida();
 }

 x= (int **)malloc(m*sizeof(int));

 if(x==NULL){
        fprintf(stderr,"?? ERROR ==> Archivo '%s' linea '%d'\n",file,linea);
	fprintf(stderr,"             Memoria Dinamica no disponible.\n");
	msalida();
 }

 for(i=0;i<m;i++)
 {
   x[i]=(int *)malloc(n*sizeof(int));
   if(x[i]==NULL){
        fprintf(stderr,"?? ERROR ==> Archivo '%s' linea '%d'\n",file,linea);
	fprintf(stderr,"             Memoria Dinamica no disponible.\n");
	msalida();
   }
 }
 return x;
}

/*---------------------------------------------------------------------------*/

double *make_dvector_lf(int n,int linea, char *file)
{
 double *x;
 if(n<0){
        fprintf(stderr,"?? ERROR ==> Archivo '%s' linea '%d'\n",file,linea);
	fprintf(stderr,"             La memoria no puede ser negativa '%d'.\n",n);
	msalida();
 }

 x= (double *)malloc(n*sizeof(double));
 if(x==NULL){
        fprintf(stderr,"?? ERROR ==> Archivo '%s' linea '%d'\n",file,linea);
	fprintf(stderr,"             Memoria Dinamica no disponible.\n");
	msalida();
 }
 return x;
}

/*---------------------------------------------------------------------------*/

float *make_svector_lf(int n,int linea, char *file)
{
 float *x;

 if(n<0){
        fprintf(stderr,"?? ERROR ==> Archivo '%s' linea '%d'\n",file,linea);
	fprintf(stderr,"             La memoria no puede ser negativa '%d'.\n",n);
	msalida();
 }


 x= (float *)malloc(n*sizeof(float));
 if(x==NULL){
        fprintf(stderr,"?? ERROR ==> Archivo '%s' linea '%d'\n",file,linea);
	fprintf(stderr,"             Memoria Dinamica no disponible.\n");
	msalida();
 }
 return x;
}

/*---------------------------------------------------------------------------*/

int  *make_ivector_lf(int n,int linea, char *file)
{
 int *x;
 if(n<0){
        fprintf(stderr,"?? ERROR ==> Archivo '%s' linea '%d'\n",file,linea);
	fprintf(stderr,"             La memoria no puede ser negativa '%d'.\n",n);
	msalida();
 }

 x= (int *)malloc(n*sizeof(int));

 if(x==NULL){
     fprintf(stderr,"?? ERROR ==> Archivo '%s' linea '%d'\n",file,linea);
	fprintf(stderr,"             Memoria Dinamica no disponible.\n");
	msalida();
 }
 return x;
}

/*---------------------------------------------------------------------------*/

void free_dmatrix_lf(double **a,int m)
{
 int i;
 for(i=0;i<m;i++)
    free(a[i]);

 free(a);
}

/*---------------------------------------------------------------------------*/

void free_smatrix_lf(float **a,int m)
{
 int i;
 for(i=0;i<m;i++)
  free(a[i]);

 free(a);
}

/*---------------------------------------------------------------------------*/

void free_imatrix_lf(int **a,int m)
{
 int i;
 for(i=0;i<m;i++)
  free(a[i]);

 free(a);
}

/*---------------------------------------------------------------------------*/

void free_dvector_lf(double *x)
{
  free(x);
}

/*---------------------------------------------------------------------------*/

void free_svector_lf(float *x)
{
  free(x);
}

/*---------------------------------------------------------------------------*/

void free_ivector_lf(int *x)
{
  free(x);
}

/*---------------------------------------------------------------------------*/

#endif //_JAG_MEM
