#include <stdio.h>
#include <math.h>

#ifdef   INTERIOR
#undef  INTERIOR
#endif

#ifdef    NODO_ACTUALIZA
#undef  NODO_ACTUALIZA
#endif

#ifdef   ESQUINA
#undef ESQUINA
#endif

#ifdef  FRONTERA_REAL
#undef FRONTERA_REAL
#endif

#ifdef  FRONTERA_ARTIFICIAL
#undef FRONTERA_ARTIFICIAL
#endif

//INTERIOR, NODO_ACTUALIZA, ESQUINA, FRONTERA_REAL, FRONTERA_ARTIFICIAL

enum  TIPO {TR,TSE,TSO,TNE,TNO}; //Izquierda,Derecha,Raiz
enum  DIRECCION {ARRIBA,ABAJO,DERECHA,IZQUIERDA};
enum  TIPO_NODO {INTERIOR, NODO_ACTUALIZA, ESQUINA, FRONTERA_REAL, FRONTERA_ARTIFICIAL};


#define  MAX_SUBDOMAINS 512

struct  tree{
 double x,y;
 double c;   // para MQ
 int    nivel;      //nivel de refinamiento
 double xmin,xmax;  //limite izquierdo y limite derecho en X
 double ymin,ymax;  //limite izquierdo y limite derecho en X
 TIPO       tipo;      //que tipo de casilla-nodo es: SE,SO,NE,NO, para el arbol
 TIPO_NODO  tipo_nodo; //para la EDP, interior, frontera, etc
 int    band;       //auxiliar para refinar todas las hojas
 double sol;        //almacena la solucion de la EDP
 double error;      //almacena el error local de interpolacion
 int    id;         //identificador unico del nodo
 int    can_remove; // 1==sip, 0 == nop
 tree  *papa;       //apuntador hacia el papa
 tree  *SE;
 tree  *SO;
 tree  *NE;
 tree  *NO;

 int   refina;
};

//-------------------------------------------------------------------
void  search(tree *papa,double xc,double yc);
void  restriccion_suavidad(tree *papa,int &cont, int max_nivel_ref);
void  marca_frontera(tree *papa);
void  cont_hojas(tree *papa,int &N);
void  get_datos_frontera(tree *papa, double *x, double *y, tree **data,int &N);
void  get_datos_frontera(tree *papa, double *x, double *y, int &N);
void  get_datos_interior(tree *papa, double *x, double *y, tree **data,int &N);
void  get_datos_interior(tree *papa, double *x, double *y, int &N);
void  get_datos(tree *papa, double *x, double *y, tree **arbol, int &N);
void  get_datos(tree *papa, double *x, double *y, int &N);
int   remueve(tree *nodo, int min_nivel_refinamiento);
int   remueve2(tree *nodo, int min_nivel_refinamiento);
int   refina(tree *nodo, int max_nivel_refinamiento);
void  refina_todo(tree *papa, int max_nivel_ref);
void  clean_band(tree *papa);
void  clean_frontera(tree *papa);
tree* get_mizq(tree *papa);
tree* get_sese(tree *papa);
tree* get_nene(tree *papa);
tree* get_nono(tree *papa);
tree* get_soso(tree *papa);
void  marca_esquinas(tree *papa);
int   isvecino(TIPO   tipo, DIRECCION  dir);
int   reflect(TIPO   tipo, DIRECCION  dir);
tree* find_vecino(tree *ptr, DIRECCION dir);
void  get_subvecino(tree *ptr, DIRECCION dir, tree *vec[16], int &N);
void  show_vecinos(tree *papa);
void  get_kvecinos(tree *select, tree *Vecinos[128], int &Kv);
void  get_kvecinos2(tree *select, tree *Vecinos[128], int &Kv);

double make_newc(double val)
{
 val = val/2.0;

 return val;
}

//-------------------------------------------------------------------
tree *crea_raiz(double xmin, double xmax, double ymin, double ymax)
{
  tree *raiz;
//Creo el nodo raiz;
  raiz = new tree;
  raiz->papa=NULL;
  raiz->tipo=TR;
  raiz->nivel=0;
  raiz->x   = 0;
  raiz->y   = 0;
  raiz->c   = 1.0;
  raiz->xmin=xmin;
  raiz->xmax=xmax;
  raiz->ymin=ymin;
  raiz->ymax=ymax;
  raiz->tipo_nodo = INTERIOR;
  raiz->can_remove = 0;
  raiz->band=1;
  raiz->SE=NULL;
  raiz->SO=NULL;
  raiz->NE=NULL;
  raiz->NO=NULL;
  return raiz;
}
//-------------------------------------------------------------------
void  ajusta_fronteras(tree *raiz,double xmin,double xmax, double ymin, double ymax)
{
  double *x,*y;
  tree   **data;
  int N,Ni;

//Primero limpio todo y despues marco la frontera y las esquinas
   clean_frontera(raiz);
   marca_frontera(raiz);
// marca_esquinas(raiz);

//Obtengo los nodos tipo frontera e interiores
   N=0; cont_hojas(raiz,N);

   x=new double[N]; y=new double[N];
   Ni=0;  get_datos_interior(raiz,x,y,Ni);
   N=Ni;  get_datos_frontera(raiz,x,y,N);

// Salvo los nodos para mostrarlos : data.m, data
   //save_nodos(x,y,Ni,N);

//Ahora lo que hago es mover las condenadas esquinas a las orillas....jejeje.
  tree *select;

 //esquina inferior-izquierda
   select = get_sese(raiz);
   select->x = xmin;  select->y=ymin;

 //esquina inferior-derecha
   select = get_soso(raiz);
   select->x = xmax;  select->y=ymin;

 //esquina superior-derecha
   select = get_nene(raiz);
   select->x = xmin;  select->y=ymax;

 //esquina superior-izquierda
   select = get_nono(raiz);
   select->x = xmax;  select->y=ymax;

//Vuelvo a obtener los datos
   data = (tree **)malloc(N*sizeof(tree));

   Ni=0;  get_datos_interior(raiz,x,y,data,Ni);
   N=Ni;  get_datos_frontera(raiz,x,y,data,N);


//Ok, hasta aqui todo va bien, falta hacerlo con las orillas del dominio
  tree *veci;
//frontera izquierda
  for(int i=Ni;i<N;i++)
  {
     veci = find_vecino(data[i],IZQUIERDA);
     if(veci==NULL)  data[i]->x = xmin;
  }
//frontera derecha
  for(int i=Ni;i<N;i++)
  {
     veci = find_vecino(data[i],DERECHA);
     if(veci==NULL)       data[i]->x = xmax;
  }
//frontera abajo
  for(int i=Ni;i<N;i++)
  {
     veci = find_vecino(data[i],ABAJO);
     if(veci==NULL)       data[i]->y = ymin;

  }
//frontera arriba
  for(int i=Ni;i<N;i++)
  {
     veci = find_vecino(data[i],ARRIBA);
     if(veci==NULL)       data[i]->y = ymax;

  }

//Salvo los nodos, primero los obtengo y despues los salvo
  //save_nodos(raiz);

//Libero la memoria
  delete[] x;
  delete[] y;
  free(data);
}
//-------------------------------------------------------------------
void  restriccion_suavidad(tree *papa,int &cont, int max_nivel_ref)
//Checo la restriccion de suavidad 2:1
{
 int i;
 if(papa==NULL)
   return;

 restriccion_suavidad(papa->SE,cont,max_nivel_ref);
 restriccion_suavidad(papa->SO,cont,max_nivel_ref);
 restriccion_suavidad(papa->NE,cont,max_nivel_ref);
 restriccion_suavidad(papa->NO,cont,max_nivel_ref);

 if(papa->SE==NULL & papa->SO==NULL & papa->NE==NULL & papa->NO==NULL)
 {
   tree *Vecinos[128];
   int   K,COTA;
   K=0;

    get_kvecinos2(papa,Vecinos,K);



  for(i=0;i<K;i++)
  {
      if(fabs(double(papa->nivel)-double(Vecinos[i]->nivel))>1.0)
      {
        if(papa->nivel>Vecinos[i]->nivel){
           if(refina(Vecinos[i],max_nivel_ref)==1)
             cont=cont+3;
        }
        else{
           if(refina(papa,max_nivel_ref)==1)
             cont=cont+3;
         }
      }
   }      // if der
}
}
//-------------------------------------------------------------------
void  marca_frontera(tree *papa)
{
 if(papa==NULL)
   return;

 marca_frontera(papa->SE);
 marca_frontera(papa->SO);
 marca_frontera(papa->NE);
 marca_frontera(papa->NO);

 if(papa->SE==NULL & papa->SO==NULL & papa->NE==NULL & papa->NO==NULL)
 {
   tree *ptr;
   ptr = find_vecino(papa,ARRIBA);
   if(ptr==NULL)
          papa->tipo_nodo = FRONTERA_REAL;  //FRONTERA_H2

   ptr = find_vecino(papa,ABAJO);
   if(ptr==NULL)
          papa->tipo_nodo = FRONTERA_REAL; //FRONTERA_H1

   ptr = find_vecino(papa,DERECHA);
   if(ptr==NULL)
          papa->tipo_nodo = FRONTERA_REAL;  //FRONTERA_V2

   ptr = find_vecino(papa,IZQUIERDA);
   if(ptr==NULL)
          papa->tipo_nodo = FRONTERA_REAL;  //FRONTERA_V1
 }
}
//-------------------------------------------------------------------
void  cont_hojas(tree *papa,int &N)
{
 if(papa==NULL)
   return;

 cont_hojas(papa->SE,N);
 cont_hojas(papa->SO,N);
 cont_hojas(papa->NE,N);
 cont_hojas(papa->NO,N);

 if(papa->SE==NULL & papa->SO==NULL & papa->NE==NULL & papa->NO==NULL)
 {
   N++;
 }
}

//-------------------------------------------------------------------
void  get_datos_frontera(tree *papa, double *x, double *y, tree **data,int &N)
{
 if(papa==NULL)
   return;

 get_datos_frontera(papa->SE,x,y,data,N);
 get_datos_frontera(papa->SO,x,y,data,N);
 get_datos_frontera(papa->NE,x,y,data,N);
 get_datos_frontera(papa->NO,x,y,data,N);

 if(papa->SE==NULL & papa->SO==NULL & papa->NE==NULL & papa->NO==NULL)
 {
   if(papa->tipo_nodo!=INTERIOR)
   {
   x[N]  = papa->x;
   y[N]  = papa->y;
   data[N] = papa;
   N++;
   }
 }
}
//-------------------------------------------------------------------
void  get_datos_frontera(tree *papa, double *x, double *y, int &N)
{
 if(papa==NULL)
   return;

 get_datos_frontera(papa->SE,x,y,N);
 get_datos_frontera(papa->SO,x,y,N);
 get_datos_frontera(papa->NE,x,y,N);
 get_datos_frontera(papa->NO,x,y,N);

 if(papa->SE==NULL & papa->SO==NULL & papa->NE==NULL & papa->NO==NULL)
 {
   if(papa->tipo_nodo!=INTERIOR)
   {
   x[N]  = papa->x;
   y[N]  = papa->y;
   N++;
   }
 }
}
//-------------------------------------------------------------------
void  get_datos_interior(tree *papa, double *x, double *y, tree **data,int &N)
{
 if(papa==NULL)
   return;

 get_datos_interior(papa->SE,x,y,data,N);
 get_datos_interior(papa->SO,x,y,data,N);
 get_datos_interior(papa->NE,x,y,data,N);
 get_datos_interior(papa->NO,x,y,data,N);

 if(papa->SE==NULL & papa->SO==NULL & papa->NE==NULL & papa->NO==NULL)
 {
   if(papa->tipo_nodo==INTERIOR)
   {
   x[N]  = papa->x;
   y[N]  = papa->y;
   data[N] = papa;
   N++;
   }
 }
}
//-------------------------------------------------------------------
void  get_datos_interior(tree *papa, double *x, double *y, int &N)
{
 if(papa==NULL)
   return;

 get_datos_interior(papa->SE,x,y,N);
 get_datos_interior(papa->SO,x,y,N);
 get_datos_interior(papa->NE,x,y,N);
 get_datos_interior(papa->NO,x,y,N);

 if(papa->SE==NULL & papa->SO==NULL & papa->NE==NULL & papa->NO==NULL)
 {
   if(papa->tipo_nodo==INTERIOR)
   {
   x[N]  = papa->x;
   y[N]  = papa->y;
   N++;
   }
 }
}
//-------------------------------------------------------------------
void  get_datos(tree *papa, double *x, double *y, tree **arbol, int &N)
{
 if(papa==NULL)
   return;

 get_datos(papa->SE,x,y,arbol,N);
 get_datos(papa->SO,x,y,arbol,N);
 get_datos(papa->NE,x,y,arbol,N);
 get_datos(papa->NO,x,y,arbol,N);

 if(papa->SE==NULL & papa->SO==NULL & papa->NE==NULL & papa->NO==NULL)
 {
   x[N]  = papa->x;
   y[N]  = papa->y;
   arbol[N] = papa;
   papa->id = N; //identificador unico
   N++;
 }
}
//-------------------------------------------------------------------
void  get_datos(tree *papa, double *x, double *y, int &N)
{
 if(papa==NULL)
   return;

 get_datos(papa->SE,x,y,N);
 get_datos(papa->SO,x,y,N);
 get_datos(papa->NE,x,y,N);
 get_datos(papa->NO,x,y,N);

 if(papa->SE==NULL & papa->SO==NULL & papa->NE==NULL & papa->NO==NULL)
 {
   x[N]  = papa->x;
   y[N]  = papa->y;
   N++;
 }
}
//-------------------------------------------------------------------
int remueve2(tree *nodo, int min_nivel_refinamiento)
//Dado un nodo marcado para refinar, lo quiero remover.
//Existen varios escenarios que hay que considerar:
// a) El nodo es nulo, en este caso no hago nada.
// b) Alguno de sus hijos no esta vacio, en esta caso no lo puedo remover
//    debido a que solo puedo remover las hojas del arbol. Donde se sabe
//    que los hijos de la hoja del arbol son nulas.
// c) Es la raiz del arbol, la cual no se quiere eliminar
// d) Paso las 3 restricciones anteriores, pero ya estoy en el limite minimo
//    de niveles de refinamiento del arbol
{
 tree *izq,*der;

//Caso  (a)
 if(nodo==NULL)
   return 0;

if(nodo->papa==NULL)
   return 0;

if(nodo->tipo==TR)
    return 0;

if(nodo->can_remove==0)
    return 0;    


//Caso  (b)
//Si alguno de sus apuntadores no esta vacio, entonces no lo refino
 if(nodo->SE!=NULL | nodo->SO!=NULL | nodo->NE!=NULL | nodo->NO!=NULL){
    return 0;
 }

 tree *papa;
 papa = nodo->papa;


//Pregunto por cada uno de sus vecinos de la celda
  if(papa->SE!=NULL){
   if(papa->SE->SE!=NULL | papa->SE->SO!=NULL | papa->SE->NE!=NULL | papa->SE->NO!=NULL){
     return 0;
   }  
 }

  if(papa->SO!=NULL){
   if(papa->SO->SE!=NULL | papa->SO->SO!=NULL | papa->SO->NE!=NULL | papa->SO->NO!=NULL){
     return 0;
    }
   }

  if(papa->NE!=NULL){
   if(papa->NE->SE!=NULL | papa->NE->SO!=NULL | papa->NE->NE!=NULL | papa->NE->NO!=NULL){
     return 0;
    }
  }

  if(papa->NO!=NULL){
   if(papa->NO->SE!=NULL | papa->NO->SO!=NULL | papa->NO->NE!=NULL | papa->NO->NO!=NULL){
     return 0;
    }
  }


//Caso  (d)
//Pregunto por el minimo nivel de refinamiento
  if(nodo->nivel>min_nivel_refinamiento)
  {
   papa = nodo->papa;
//Borro el lado derecho
  delete[]  papa->SE;
  delete[]  papa->SO;
  delete[]  papa->NE;
  delete[]  papa->NO;

  papa->SE = NULL;
  papa->SO = NULL;
  papa->NE = NULL;
  papa->NO = NULL;


  return 1;
  }
  else
    return 0;
}
//-------------------------------------------------------------------




//-------------------------------------------------------------------
int remueve(tree *nodo, int min_nivel_refinamiento)
//Dado un nodo marcado para refinar, lo quiero remover.
//Existen varios escenarios que hay que considerar:
// a) El nodo es nulo, en este caso no hago nada.
// b) Alguno de sus hijos no esta vacio, en esta caso no lo puedo remover
//    debido a que solo puedo remover las hojas del arbol. Donde se sabe
//    que los hijos de la hoja del arbol son nulas.
// c) Es la raiz del arbol, la cual no se quiere eliminar
// d) Paso las 3 restricciones anteriores, pero ya estoy en el limite minimo
//    de niveles de refinamiento del arbol
{
 tree *izq,*der;

//Caso  (a)
 if(nodo==NULL)
   return 0;

if(nodo->papa==NULL)
   return 0;

if(nodo->tipo==TR)
    return 0;

//Si es un nodo de tipo esquina, entonces no lo remuevo
 if(nodo->tipo_nodo==ESQUINA)
    return 0;

//Caso  (b)
//Si alguno de sus apuntadores no esta vacio, entonces no lo refino
 if(nodo->SE!=NULL | nodo->SO!=NULL | nodo->NE!=NULL | nodo->NO!=NULL){
    return 0;
 }

 tree *papa;
 papa = nodo->papa;

// if(papa->tipo_nodo==ESQUINA){ // ojo, no se bien porque
//    return 0;
// }


//Pregunto por cada uno de sus vecinos de la celda
  if(papa->SE!=NULL){
   if(papa->SE->SE!=NULL | papa->SE->SO!=NULL | papa->SE->NE!=NULL | papa->SE->NO!=NULL){
     return 0;
   }
 }

  if(papa->SO!=NULL){
   if(papa->SO->SE!=NULL | papa->SO->SO!=NULL | papa->SO->NE!=NULL | papa->SO->NO!=NULL){
     return 0;
    }
   }

  if(papa->NE!=NULL){
   if(papa->NE->SE!=NULL | papa->NE->SO!=NULL | papa->NE->NE!=NULL | papa->NE->NO!=NULL){
     return 0;
    }
  }

  if(papa->NO!=NULL){
   if(papa->NO->SE!=NULL | papa->NO->SO!=NULL | papa->NO->NE!=NULL | papa->NO->NO!=NULL){
     return 0;
    }
  }

//Pregunto si algun vecino del mismo cuadrante es de tipo frontera
 if(papa->SE!=NULL){
  if(papa->SE->tipo_nodo!=INTERIOR){
     return 0;
  }
 }
 if(papa->SO!=NULL){
   if(papa->SO->tipo_nodo!=INTERIOR ){
        return 0;
   }
 }
 if(papa->NE!=NULL){
  if(papa->NE->tipo_nodo!=INTERIOR){
     return 0;
   }
  }

 if(papa->NO!=NULL){
   if(papa->NO->tipo_nodo!=INTERIOR ){
     return 0;
   }
 }

if(nodo->can_remove==0)
    return 0;

//Caso  (d)
//Pregunto por el minimo nivel de refinamiento
  if(nodo->nivel>min_nivel_refinamiento)
  {
   papa = nodo->papa;
//Borro el lado derecho
  delete[]  papa->SE;
  delete[]  papa->SO;
  delete[]  papa->NE;
  delete[]  papa->NO;

  papa->SE = NULL;
  papa->SO = NULL;
  papa->NE = NULL;
  papa->NO = NULL;


  return 1;
  }

 return 0;
}

//-------------------------------------------------------------------
int refina(tree *nodo, int max_nivel_refinamiento)
{
 tree *SE,*SO,*NE,*NO;

 if(nodo==NULL)
    return 0;

//Si alguno de sus apuntadores no esta vacio, entonces no lo refino
 if(nodo->SE!=NULL | nodo->SO!=NULL | nodo->NE!=NULL | nodo->NO!=NULL){
    return 0;
 }

//Veo si satisface la restriccion del maximo nivel de refinamiento
 if(nodo->nivel>= max_nivel_refinamiento) return 0;


  SE = new tree;
  SO = new tree;
  NE = new tree;
  NO = new tree;


  SE->papa   = nodo;
  SE->tipo   = TSE;
  SE->tipo_nodo = nodo->tipo_nodo;
  SE->nivel  = nodo->nivel+1;
  SE->xmin   = nodo->xmin;
  SE->xmax   = nodo->xmin + 0.5*(nodo->xmax-nodo->xmin);
  SE->ymin   = nodo->ymin;
  SE->ymax   = nodo->ymin + 0.5*(nodo->ymax-nodo->ymin);
  SE->c      = make_newc(nodo->c);
  SE->band   = 1;
  SE->can_remove = 1;
  SE->SE     = NULL;
  SE->SO     = NULL;
  SE->NE     = NULL;
  SE->NO     = NULL;
  SE->x      = SE->xmin + 0.5*(SE->xmax-SE->xmin);
  SE->y      = SE->ymin + 0.5*(SE->ymax-SE->ymin);


  SO->papa   = nodo;
  SO->tipo   = TSO;
  SO->tipo_nodo = nodo->tipo_nodo;
  SO->nivel  = nodo->nivel+1;
  SO->xmin   = SE->xmax;
  SO->xmax   = nodo->xmax;
  SO->ymin   = nodo->ymin;
  SO->ymax   = SE->ymax;
  SO->c      = make_newc(nodo->c);
  SO->band   = 1;
  SO->can_remove = 1;
  SO->SE     = NULL;
  SO->SO     = NULL;
  SO->NE     = NULL;
  SO->NO     = NULL;
  SO->x      = SO->xmin + 0.5*(SO->xmax-SO->xmin);
  SO->y      = SO->ymin + 0.5*(SO->ymax-SO->ymin);

  NE->papa   = nodo;
  NE->tipo   = TNE;
  NE->tipo_nodo = nodo->tipo_nodo;
  NE->nivel  = nodo->nivel+1;
  NE->xmin   = SE->xmin;
  NE->xmax   = SE->xmax;
  NE->ymin   = SE->ymax;
  NE->ymax   = nodo->ymax;
  NE->c      = make_newc(nodo->c);
  NE->band   = 1;
  NE->can_remove = 1;
  NE->SE     = NULL;
  NE->SO     = NULL;
  NE->NE     = NULL;
  NE->NO     = NULL;
  NE->x      = NE->xmin + 0.5*(NE->xmax-NE->xmin);
  NE->y      = NE->ymin + 0.5*(NE->ymax-NE->ymin);

  NO->papa   = nodo;
  NO->tipo   = TNO;
  NO->tipo_nodo = nodo->tipo_nodo;
  NO->nivel  = nodo->nivel+1;
  NO->xmin   = SO->xmin;
  NO->xmax   = SO->xmax;
  NO->ymin   = SO->ymax;
  NO->ymax   = nodo->ymax;
  NO->c      = make_newc(nodo->c);
  NO->band   = 1;
  NO->can_remove = 1;
  NO->SE     = NULL;
  NO->SO     = NULL;
  NO->NE     = NULL;
  NO->NO     = NULL;
  NO->x      = NO->xmin + 0.5*(NO->xmax-NO->xmin);
  NO->y      = NO->ymin + 0.5*(NO->ymax-NO->ymin);


  nodo->SE = SE;
  nodo->SO = SO;
  nodo->NE = NE;
  nodo->NO = NO;

 return 1;
}


//-------------------------------------------------------------------
void refina_nodo(tree *nodo, double *X, double *Y)
{
 tree *SE,*SO,*NE,*NO;

  SE = new tree;
  SO = new tree;
  NE = new tree;
  NO = new tree;


  SE->xmin   = nodo->xmin;
  SE->xmax   = nodo->xmin + 0.5*(nodo->xmax-nodo->xmin);
  SE->ymin   = nodo->ymin;
  SE->ymax   = nodo->ymin + 0.5*(nodo->ymax-nodo->ymin);
  SE->c      = make_newc(nodo->c);
  SE->x      = SE->xmin + 0.5*(SE->xmax-SE->xmin);
  SE->y      = SE->ymin + 0.5*(SE->ymax-SE->ymin);


  SO->xmin   = SE->xmax;
  SO->xmax   = nodo->xmax;
  SO->ymin   = nodo->ymin;
  SO->ymax   = SE->ymax;
  SO->x      = SO->xmin + 0.5*(SO->xmax-SO->xmin);
  SO->y      = SO->ymin + 0.5*(SO->ymax-SO->ymin);

  NE->xmin   = SE->xmin;
  NE->xmax   = SE->xmax;
  NE->ymin   = SE->ymax;
  NE->ymax   = nodo->ymax;
  NE->x      = NE->xmin + 0.5*(NE->xmax-NE->xmin);
  NE->y      = NE->ymin + 0.5*(NE->ymax-NE->ymin);

  NO->xmin   = SO->xmin;
  NO->xmax   = SO->xmax;
  NO->ymin   = SO->ymax;
  NO->ymax   = nodo->ymax;
  NO->x      = NO->xmin + 0.5*(NO->xmax-NO->xmin);
  NO->y      = NO->ymin + 0.5*(NO->ymax-NO->ymin);


  X[0] = SE->x; Y[0] = SE->y;
  X[1] = SO->x; Y[1] = SO->y;
  X[2] = NE->x; Y[2] = NE->y;
  X[3] = NO->x; Y[3] = NO->y;

  delete SE;
  delete SO;
  delete NE;
  delete NO;

}

//-------------------------------------------------------------------
void refina_nodo_esquina(tree *nodo, double *X, double *Y)
{
 tree *SE,*SO,*NE,*NO;

  SE = new tree;
  SO = new tree;
  NE = new tree;
  NO = new tree;

// NE NO
// SE SO
  SE->x      = nodo->xmin + 0.5*(nodo->xmax-nodo->xmin);
  SE->y      = nodo->ymin;

  NE->x      = nodo->xmin;
  NE->y      = nodo->ymin + 0.5*(nodo->ymax-nodo->ymin);


// NE NO
// SE SO
  X[0] = SE->x; Y[0] = SE->y;
  X[1] = NE->x; Y[2] = NE->y;


  delete SE;
  delete SO;
  delete NE;
  delete NO;

}

//-------------------------------------------------------------------

//-------------------------------------------------------------------
void refina_todo(tree *papa, int max_nivel_ref)
{

 if(papa==NULL)
   return;

 refina_todo(papa->SE,max_nivel_ref);
 refina_todo(papa->SO,max_nivel_ref);
 refina_todo(papa->NE,max_nivel_ref);
 refina_todo(papa->NO,max_nivel_ref);


 if(papa->SE==NULL & papa->SO==NULL & papa->NE==NULL & papa->NO==NULL)
 if(papa->band==1)
 {
    if(refina(papa,max_nivel_ref)==1)
    {
    if(papa->SE!=NULL)  papa->SE->band=0;
    if(papa->SO!=NULL)  papa->SO->band=0;
    if(papa->NE!=NULL) papa->NE->band=0;
    if(papa->NO!=NULL) papa->NO->band=0;
   }
 }
}
//-------------------------------------------------------------------
void clean_band(tree *papa)
{

 if(papa==NULL)
   return;

 clean_band(papa->SE);
 clean_band(papa->SO);
 clean_band(papa->NE);
 clean_band(papa->NO);

 if(papa->SE==NULL & papa->SO==NULL & papa->NE==NULL & papa->NO==NULL)
 {
    papa->band=1;
 }
}
//-------------------------------------------------------------------
void clean_frontera(tree *papa)
{

 if(papa==NULL)
   return;

 clean_frontera(papa->SE);
 clean_frontera(papa->SO);
 clean_frontera(papa->NE);
 clean_frontera(papa->NO);

 if(papa->SE==NULL & papa->SO==NULL & papa->NE==NULL & papa->NO==NULL)
 {
    papa->tipo_nodo=INTERIOR;
 }
}
//-------------------------------------------------------------------
tree* get_mizq(tree *papa)
{
 if(papa==NULL)
   return NULL;

 if(papa->SE==NULL & papa->SO==NULL & papa->NE==NULL & papa->NO==NULL)
    return papa;

 get_mizq(papa->SE);
}
//-------------------------------------------------------------------
tree* get_sese(tree *papa)
{
 if(papa==NULL)   return NULL;
 if(papa->SE==NULL & papa->SO==NULL & papa->NE==NULL & papa->NO==NULL)    return papa;
 get_sese(papa->SE);
}
//-------------------------------------------------------------------
tree* get_nene(tree *papa)
{
 if(papa==NULL)   return NULL;
 if(papa->SE==NULL & papa->SO==NULL & papa->NE==NULL & papa->NO==NULL)    return papa;
 get_nene(papa->NE);
}
//-------------------------------------------------------------------
tree* get_nono(tree *papa)
{
 if(papa==NULL)   return NULL;
 if(papa->SE==NULL & papa->SO==NULL & papa->NE==NULL & papa->NO==NULL)    return papa;
 get_nono(papa->NO);
}
//-------------------------------------------------------------------
tree* get_soso(tree *papa)
{
 if(papa==NULL)   return NULL;
 if(papa->SE==NULL & papa->SO==NULL & papa->NE==NULL & papa->NO==NULL)    return papa;
 get_soso(papa->SO);
}
//-------------------------------------------------------------------
void marca_esquinas(tree *papa)
{
  tree *ptr;
  ptr = get_sese(papa); ptr->tipo_nodo = ESQUINA;
  ptr = get_nene(papa); ptr->tipo_nodo = ESQUINA;
  ptr = get_nono(papa); ptr->tipo_nodo = ESQUINA;
  ptr = get_soso(papa); ptr->tipo_nodo = ESQUINA;
}
//-------------------------------------------------------------------
int isvecino(TIPO   tipo, DIRECCION  dir)
{
 switch(tipo)
 {
  case TNE:
             switch(dir)
             {
               case  ARRIBA:
               case  IZQUIERDA:  return 0;
                               break;
               case  ABAJO:
               case  DERECHA:  return 1;
                               break;
             }
        break;
  case TNO:
             switch(dir)
             {
               case  ARRIBA:
               case  DERECHA:  return 0;
                               break;
               case  ABAJO:
               case  IZQUIERDA: return 1;
                               break;
             }
        break;
  case TSE:
             switch(dir)
             {
               case  ARRIBA:
               case  DERECHA:  return 1;
                               break;
               case  ABAJO:
               case  IZQUIERDA: return 0;
                               break;
             }
        break;

  case TSO:
             switch(dir)
             {
               case  ARRIBA:
               case  IZQUIERDA:  return 1;
                               break;
               case  DERECHA:
               case  ABAJO: return 0;
                               break;
             }
        break;
 }
 return -1;
}
//-------------------------------------------------------------------
int reflect(TIPO   tipo, DIRECCION  dir)
{
 switch(tipo)
 {
  case TNE:
             switch(dir)
             {
               case  ARRIBA:     return TSE;
               case  DERECHA:    return TNO;
               case  ABAJO:      return TSE;
               case  IZQUIERDA:  return TNO;
             }
        break;
  case TNO:
             switch(dir)
             {
               case  ARRIBA:     return TSO;
               case  DERECHA:    return TNE;
               case  ABAJO:      return TSO;
               case  IZQUIERDA:  return TNE;
             }
        break;
  case TSE:
             switch(dir)
             {
               case  ARRIBA:     return TNE;
               case  DERECHA:    return TSO;
               case  ABAJO:      return TNE;
               case  IZQUIERDA:  return TSO;
             }
        break;
  case TSO:
             switch(dir)
             {
               case  ARRIBA:     return TNO;
               case  DERECHA:    return TSE;
               case  ABAJO:      return TNO;
               case  IZQUIERDA:  return TSE;
             }
        break;
 }
}

//-------------------------------------------------------------------
tree * find_vecino(tree *ptr, DIRECCION dir)
{
 static tree *aux;
 int tipo;

 if(ptr==NULL){
   aux=NULL;
   return NULL;
 }

 if(ptr->tipo==TR){
    aux=NULL;
    return NULL;
}

 if(isvecino(ptr->tipo,dir)==0){
     find_vecino(ptr->papa,dir);
 }
 else{
     //switch
          tipo = reflect(ptr->tipo,dir);
          aux  = ptr->papa;
          switch(tipo)
          {
           case TSE:  aux = aux->SE; break;
           case TSO:  aux = aux->SO; break;
           case TNE:  aux = aux->NE; break;
           case TNO:  aux = aux->NO; break;
          }
         return aux;
 }
 {   //desciendo
       if(aux==NULL)
          return NULL;

          tipo = reflect(ptr->tipo,dir);
          switch(tipo)
          {
           case TSE:
                     if(aux->SE!=NULL)
                         aux = aux->SE;
                     break;
           case TSO:
                     if(aux->SO!=NULL)
                        aux = aux->SO;
                     break;
           case TNE:
                     if(aux->NE!=NULL)
                       aux = aux->NE;
                     break;
           case TNO:
                     if(aux->NO!=NULL)
                        aux = aux->NO;
                     break;
          }
          return aux;
 }
}
//-------------------------------------------------------------------
void get_subvecino(tree *ptr, DIRECCION dir, tree *vec[16], int &N)
{
 int ip1,ip2;
 int tipo;

 if(ptr==NULL){
   return;
 }

//Primero checo si  existen subvecinos
  if(ptr->SE==NULL | ptr->SO==NULL | ptr->NE==NULL | ptr->NO==NULL)
     return;

//Obtengo los nodos al primer nivel
  switch(dir)
  {
   case  ARRIBA:
                 if(ptr->SE->SE==NULL & ptr->SE->SO==NULL & ptr->SE->NE==NULL & ptr->SE->NO==NULL)
                    vec[N++] = ptr->SE;

                 if(ptr->SO->SE==NULL & ptr->SO->SO==NULL & ptr->SO->NE==NULL & ptr->SO->NO==NULL)
                  vec[N++] = ptr->SO;

                 get_subvecino(ptr->SE,dir,vec,N);
                 get_subvecino(ptr->SO,dir,vec,N);
                 break;
   case  ABAJO:
                 if(ptr->NE->SE==NULL & ptr->NE->SO==NULL & ptr->NE->NE==NULL & ptr->NE->NO==NULL)
                    vec[N++] = ptr->NE;

                 if(ptr->NO->SE==NULL & ptr->NO->SO==NULL & ptr->NO->NE==NULL & ptr->NO->NO==NULL)
                  vec[N++] = ptr->NO;

                 get_subvecino(ptr->NE,dir,vec,N);
                 get_subvecino(ptr->NO,dir,vec,N);
                 break;
   case  DERECHA:
                 if(ptr->NE->SE==NULL & ptr->NE->SO==NULL & ptr->NE->NE==NULL & ptr->NE->NO==NULL)
                    vec[N++] = ptr->NE;

                 if(ptr->SE->SE==NULL & ptr->SE->SO==NULL & ptr->SE->NE==NULL & ptr->SE->NO==NULL)
                    vec[N++] = ptr->SE;

                 get_subvecino(ptr->SE,dir,vec,N);
                 get_subvecino(ptr->NE,dir,vec,N);
                 break;
   case  IZQUIERDA:
                 if(ptr->NO->SE==NULL & ptr->NO->SO==NULL & ptr->NO->NE==NULL & ptr->NO->NO==NULL)
                  vec[N++] = ptr->NO;

                 if(ptr->SO->SE==NULL & ptr->SO->SO==NULL & ptr->SO->NE==NULL & ptr->SO->NO==NULL)
                  vec[N++] = ptr->SO;

                 get_subvecino(ptr->SO,dir,vec,N);
                 get_subvecino(ptr->NO,dir,vec,N);
                 break;
  }

/*  
  get_subvecino(ptr->SE,dir,vec,N);
  get_subvecino(ptr->SO,dir,vec,N);
  get_subvecino(ptr->NE,dir,vec,N);
  get_subvecino(ptr->NO,dir,vec,N);
*/

}
//-------------------------------------------------------------------
void show_vecinos(tree *papa)
{

 if(papa==NULL)
   return;

 show_vecinos(papa->SE);
 show_vecinos(papa->SO);
 show_vecinos(papa->NE);
 show_vecinos(papa->NO);

 if(papa->SE==NULL & papa->SO==NULL & papa->NE==NULL & papa->NO==NULL)
 {
// Busco el vecino ...jeje
   tree *select;
   tree *ptr;
   tree *vecinos[128];
   int   N;

   select = papa;

    printf("------------------------------------\n");
    printf("nodo = [%d] %f %f\n",select->nivel,select->x,select->y);
// Vecino de arriba
   ptr = find_vecino(select,ARRIBA);
   if(ptr!=NULL){
     N=0;
     get_subvecino(ptr,ARRIBA,vecinos,N);
     if(N==0){ vecinos[0]=ptr; N=1;}

    printf("arriba    = ");
    for(int i=0;i<N;i++){
    if(i==0)
        printf("[%d] %f %f\n",vecinos[i]->nivel,vecinos[i]->x,vecinos[i]->y);
     if(i>0)
       printf("            [%d] %f %f\n",vecinos[i]->nivel,vecinos[i]->x,vecinos[i]->y);
      }
   }

// Vecino de abajo
   ptr = find_vecino(select,ABAJO);
   if(ptr!=NULL){
     N=0;
     get_subvecino(ptr,ABAJO,vecinos,N);
     if(N==0){ vecinos[0]=ptr; N=1; }

    printf("abajo     = ");
    for(int i=0;i<N;i++){
    if(i==0)
        printf("[%d] %f %f\n",vecinos[i]->nivel,vecinos[i]->x,vecinos[i]->y);
     if(i>0)
       printf("            [%d] %f %f\n",vecinos[i]->nivel,vecinos[i]->x,vecinos[i]->y);
   }
   }

// Vecino de derecha
   ptr = find_vecino(select,DERECHA);
   if(ptr!=NULL){
     N=0;
     get_subvecino(ptr,DERECHA,vecinos,N);
     if(N==0){ vecinos[0]=ptr; N=1;}

    printf("derecha   = ");
    for(int i=0;i<N;i++){
    if(i==0)
      printf("[%d] %f %f\n",vecinos[i]->nivel,vecinos[i]->x,vecinos[i]->y);
     if(i>0)
       printf("            [%d] %f %f\n",vecinos[i]->nivel,vecinos[i]->x,vecinos[i]->y);

    }
   }

// Vecino de izquierda
   ptr = find_vecino(select,IZQUIERDA);
   if(ptr!=NULL){
     N=0;
     get_subvecino(ptr,IZQUIERDA,vecinos,N);
     if(N==0){ vecinos[0]=ptr; N=1;}

    printf("izquierda = ");
    for(int i=0;i<N;i++){
    if(i==0)
        printf("[%d] %f %f\n",vecinos[i]->nivel,vecinos[i]->x,vecinos[i]->y);
     if(i>0)
       printf("            [%d] %f %f\n",vecinos[i]->nivel,vecinos[i]->x,vecinos[i]->y);
   }
   }

 }
}
//-------------------------------------------------------------------
void get_kvecinos(tree *select, tree *Vecinos[128], int &Kv)
{
  tree *vecinos[32];
  tree *ptr;
  int  i,m,K;
  K=0;

//Vecino de arriba
  ptr = find_vecino(select,ARRIBA);
  if(ptr!=NULL){
     m=0;
     get_subvecino(ptr,ARRIBA,vecinos,m);
     if(m==0){ vecinos[0]=ptr;m=1;}
     for(i=0;i<m;i++)
     {
      if(vecinos[i]!=NULL)
      {
       Vecinos[K]=vecinos[i]; K++;
      }
     }
  }
// Vecino de abajo
   ptr = find_vecino(select,ABAJO);
   if(ptr!=NULL){
      m=0;
      get_subvecino(ptr,ABAJO,vecinos,m);
     if(m==0){ vecinos[0]=ptr; m=1; }
     for(i=0;i<m;i++)
     {
      if(vecinos[i]!=NULL)
      {
       Vecinos[K]=vecinos[i]; K++;
      }
     }
   }
// Vecino de derecha
   ptr = find_vecino(select,DERECHA);
   if(ptr!=NULL){
    m=0;
    get_subvecino(ptr,DERECHA,vecinos,m);
    if(m==0){ vecinos[0]=ptr;m=1;}
     for(i=0;i<m;i++)
     {
      if(vecinos[i]!=NULL)
      {
       Vecinos[K]=vecinos[i]; K++;
      }
     }
   }
//Determino el vecino de derecha-arriba
//Determino el vecino de derecha-abajo

 // Vecino de izquierda
    ptr = find_vecino(select,IZQUIERDA);
    if(ptr!=NULL){
    m=0;
    get_subvecino(ptr,IZQUIERDA,vecinos,m);
    if(m==0){ vecinos[0]=ptr; m=1;}
     for(i=0;i<m;i++)
     {
      if(vecinos[i]!=NULL)
      {
       Vecinos[K]=vecinos[i]; K++;
      }
     }
   }
  Kv=K;
}


//-------------------------------------------------------------------

int is_inbox(double xmin, double xmax, double ymin, double ymax, tree *ptr)
{
 if(ptr==NULL) return 0;

//primer  punto    ptr->xmin    ptr->ymin
  if( ptr->xmin>=xmin & ptr->xmin<=xmax)
   if( ptr->ymin>=ymin & ptr->ymin<=ymax)
     return 1;
//segundo punto    ptr->xmin    ptr->ymax
  if( ptr->xmin>=xmin & ptr->xmin<=xmax)
   if( ptr->ymax>=ymin & ptr->ymax<=ymax)
     return 1;
//tercer  punto    ptr->xmax    ptr->ymin
  if( ptr->xmax>=xmin & ptr->xmax<=xmax)
   if( ptr->ymin>=ymin & ptr->ymin<=ymax)
     return 1;
//cuarto  punto    ptr->xmax    ptr->ymax
  if( ptr->xmax>=xmin & ptr->xmax<=xmax)
   if( ptr->ymax>=ymin & ptr->ymax<=ymax)
     return 1;

 return 0;
}
//-------------------------------------------------------------------
int is_in(tree **Vecinos, int cont, tree *ptr)
{
 if(cont<=0) return 0;

 for(int i=0;i<cont;i++)
 {
   if(Vecinos[i]==ptr)
   {
     return 1;
   }
 }
 return 0;
}

//-------------------------------------------------------------------
void get_kvecinos2(tree *select, tree *Vecinos[128], int &Kv)
// Encuentro los 4 vecinos naturales, por cada vecino natural, encuentro sus 4
// vecinos naturales y despues me quedo con los nodos que comparten una misma
// cara o coinciden en las esquinas
{
  tree *vecinos[128];
  tree *mas_vecinos[128];
  tree *todos_vecinos[128];
  tree *ptr;
  int   m,K,k;
  int   cont=0;
  int   cont2=0;

  K=0;   get_kvecinos(select,vecinos,K);

//Los agrego
  for(int i=0;i<K;i++)     todos_vecinos[cont++]=vecinos[i];

//Busco sus vecinos
  for(int i=0;i<K;i++)
  {
    k=0;    get_kvecinos(vecinos[i],mas_vecinos,k);

    for(int j=0;j<k;j++)
    {
     if(mas_vecinos[j]!=select)
     {
       todos_vecinos[cont++]=mas_vecinos[j];
     }
    }
 }

  double xmin,xmax,ymin,ymax;

  xmin = select->xmin;
  xmax = select->xmax;
  ymin = select->ymin;
  ymax = select->ymax;


  for(int i=0;i<cont;i++)
  {
   //pregunto por cualquiera de las esquinas de los nodos seleccionados
      if(is_inbox(xmin,xmax,ymin,ymax,todos_vecinos[i])==1)
      if(is_in(Vecinos,cont2,todos_vecinos[i])==0)
      {
        Vecinos[cont2++]=todos_vecinos[i];
      }
  }

  Kv=cont2;
}
//-------------------------------------------------------------------


int make_formato_metis(tree *nodo,int &id1, int &id2, int &id3, int &id4)
/*
1  2   3   4
5  6   7   8
9  10  11  12

Mesh
6 4
5 1 2 6
6 2 3 7
7 3 4 8
9 5 6 10
10 6 7 11
11 7 8 12

*/
{
if(nodo==NULL) return 0;

if(nodo->tipo!=TSE) return 0;

 id1 = nodo->id;
 id2 = nodo->papa->NE->id;
 id3 = nodo->papa->NO->id;
 id4 = nodo->papa->SO->id;

printf("%d %d %d %d\n",id1,id2,id3,id4);

return 1;
}

